@@ -4,11 +4,14 @@
 ## Description: collocation db, profiling relation: co-frequency database (using pair of DiaColloDB::PackedFile)
 
 package DiaColloDB::Relation::Cofreqs;
+use DiaColloDB::Compat;
 use DiaColloDB::Relation;
 use DiaColloDB::PackedFile;
 use DiaColloDB::PackedFile::MMap;
 use DiaColloDB::Utils qw(:fcntl :env :run :json :pack);
 use Fcntl qw(:DEFAULT :seek);
+use File::Basename qw(dirname);
+use version;
 use strict;
 
 ##==============================================================================
@@ -27,23 +30,28 @@ our $PFCLASS = 'DiaColloDB::PackedFile::MMap';
 ##   (
 ##    ##-- user options
 ##    class    => $class,      ##-- optional, useful for debugging from header file
-##    base     => $basename,   ##-- file basename (default=undef:none); use files "${base}.dba1", "${base}.dba2", "${base}.hdr"
+##    base     => $basename,   ##-- file basename (default=undef:none); use files "${base}.dba1", "${base}.dba2", "${base}.dba3", "${base}.hdr"
 ##    flags    => $flags,      ##-- fcntl flags or open-mode (default='r')
 ##    perms    => $perms,      ##-- creation permissions (default=(0666 &~umask))
 ##    dmax     => $dmax,       ##-- maximum distance for co-occurrences (default=5)
 ##    fmin     => $fmin,       ##-- minimum pair frequency (default=0)
 ##    pack_i   => $pack_i,     ##-- pack-template for IDs (default='N')
-##    pack_f   => $pack_f,     ##-- pack-template for IDs (default='N')
+##    pack_f   => $pack_f,     ##-- pack-template for frequencies (default='N')
+##    pack_d   => $pack_d,     ##-- pack-tempalte for dates (default='n')
 ##    keeptmp  => $bool,       ##-- keep temporary files? (default=false)
+##    logCompat => $level,     ##-- log-level for compatibility warnings (default='warn')
 ##    ##
 ##    ##-- size info (after open() or load())
 ##    size1    => $size1,      ##-- == $r1->size()
 ##    size2    => $size2,      ##-- == $r2->size()
+##    size3    => $size3,      ##-- == $r3->size()
 ##    ##
 ##    ##-- low-level data
-##    r1 => $r1,               ##-- pf: [$end2,$f1] @ $i1
-##    r2 => $r2,               ##-- pf: [$i2,$f12]  @ end2($i1-1)..(end2($i1)-1)
+##    r1 => $r1,               ##-- pf: [$end2]            @ $i1				: constant (logical index)
+##    r2 => $r2,               ##-- pf: [$end3,$d1,$f1]*   @ end2($i1-1)..(end2($i1+1)-1)	: sorted by $d1 for each $i1
+##    r3 => $r3,               ##-- pf: [$i2,$f12]*        @ end3($d1-1)..(end3($d1+1)-1)	: sorted by $i2 for each ($i1,$d1)
 ##    N  => $N,                ##-- sum($f12)
+##    version => $version,     ##-- file version, for compatibility checks
 ##   )
 sub new {
   my $that = shift;
@@ -55,11 +63,13 @@ sub new {
 		    fmin  =>0,
 		    pack_i=>'N',
 		    pack_f=>'N',
-		    r1 => $PFCLASS->new(),
-		    r2 => $PFCLASS->new(),
+		    pack_d=>'n',
 		    N  => 0,
+		    version => $DiaColloDB::VERSION,
+		    logCompat => 'warn',
 		    @_
 		   }, (ref($that)||$that));
+  $cof->{$_} //= $cof->mmclass($PFCLASS)->new() foreach (qw(r1 r2 r3));
   $cof->{class} = ref($cof);
   return $cof->open() if (defined($cof->{base}));
   return $cof;
@@ -85,19 +95,35 @@ sub open {
   $cof->close() if ($cof->opened);
   $cof->{base}  = $base;
   $cof->{flags} = $flags = fcflags($flags);
+  my ($hdr); ##-- save header, for version-checking
   if (fcread($flags) && !fctrunc($flags)) {
-    $cof->loadHeader()
-      or $cof->logconess("failed to load header from '$cof->{base}.hdr': $!");
+    $hdr = $cof->readHeader()
+      or $cof->logconess("failed to read header data from '$cof->{base}.hdr': $!");
+    $cof->loadHeaderData($hdr)
+      or $cof->logconess("failed to load header data from '$cof->{base}.hdr': $!");
   }
-  $cof->{r1}->open("$base.dba1", $flags, perms=>$cof->{perms}, packas=>"$cof->{pack_i}$cof->{pack_f}")
+
+  ##-- check compatibility
+  my $min_version = qv(0.10.000);
+  if ($hdr && (!defined($hdr->{version}) || version->parse($hdr->{version}) < $min_version)) {
+    $cof->vlog($cof->{logCompat}, "using compatibility mode for $cof->{base}.*; consider running \`dcdb-upgrade.perl ", dirname($cof->{base}), "\'");
+    DiaColloDB::Compat->usecompat('v0_09');
+    bless($cof, 'DiaColloDB::Compat::v0_09::Relation::Cofreqs');
+    $cof->{version} = $hdr->{version};
+    return $cof->open($base,$flags);
+  }
+
+  ##-- open low-level data structures
+  $cof->{r1}->open("$base.dba1", $flags, perms=>$cof->{perms}, packas=>"$cof->{pack_i}")
     or $cof->logconfess("open failed for $base.dba1: $!");
-  $cof->{r2}->open("$base.dba2", $flags, perms=>$cof->{perms}, packas=>"$cof->{pack_i}$cof->{pack_f}")
+  $cof->{r2}->open("$base.dba2", $flags, perms=>$cof->{perms}, packas=>"$cof->{pack_i}$cof->{pack_d}$cof->{pack_f}")
     or $cof->logconfess("open failed for $base.dba2: $!");
+  $cof->{r3}->open("$base.dba3", $flags, perms=>$cof->{perms}, packas=>"$cof->{pack_i}$cof->{pack_f}")
+    or $cof->logconfess("open failed for $base.dba3: $!");
   $cof->{size1} = $cof->{r1}->size;
   $cof->{size2} = $cof->{r2}->size;
+  $cof->{size3} = $cof->{r3}->size;
 
-  #$cof->debug("open(): opened level-1 relation $cof->{r1}{file}.* as ", ref($cof->{r1}));
-  #$cof->debug("open(): opened level-2 relation $cof->{r2}{file}.* as ", ref($cof->{r2}));
   return $cof;
 }
 
@@ -109,6 +135,7 @@ sub close {
   }
   $cof->{r1}->close() or return undef;
   $cof->{r2}->close() or return undef;
+  $cof->{r3}->close() or return undef;
   undef $cof->{base};
   return $cof;
 }
@@ -119,7 +146,9 @@ sub opened {
   return
     (defined($cof->{base})
      && defined($cof->{r1}) && $cof->{r1}->opened
-     && defined($cof->{r2}) && $cof->{r2}->opened);
+     && defined($cof->{r2}) && $cof->{r2}->opened
+     && defined($cof->{r3}) && $cof->{r3}->opened
+    );
 }
 
 ##--------------------------------------------------------------
@@ -129,7 +158,7 @@ sub opened {
 ## @keys = $cof->headerKeys()
 ##  + keys to save as header
 sub headerKeys {
-  return grep {!ref($_[0]{$_}) && $_ !~ m{^(?:base|flags|perms)$}} keys %{$_[0]};
+  return grep {!ref($_[0]{$_}) && $_ !~ m{^(?:base|flags|perms|log.*)$}} keys %{$_[0]};
 }
 
 ## $bool = $cof->loadHeaderData($hdr)
@@ -158,10 +187,14 @@ sub loadHeaderData {
 ##  + INHERITED from DiaColloDB::Persistent
 
 ## $cof = $cof->loadTextFh($fh,%opts)
-##  + loads from text file as saved by saveTextFh()
-##  + supports semi-sorted input: input fh must be sorted by $i1,
-##    and all $i2 for each $i1 must be adjacent (i.e. no intervening $j1 != $i1)
-##  + supports multiple lines for pairs ($i1,$i2) provided the above conditions hold
+##  + loads from text file as saved by saveTextFh():
+##      N                       ##-- 1 field : N
+##      FREQ ID1 DATE           ##-- 3 fields: un-collocated portion of $f1
+##      FREQ ID1 DATE ID2       ##-- 4 fields: co-frequency pair (ID2 >= 0)
+##      FREQ ID1 DATE ID2 DATE2 ##-- 5 fields: redundant date (used by create(); DATE2 is ignored)
+##  + supports semi-sorted input: input fh must be sorted by $i1,$d1
+##    and all $i2 for each $i1,$d1 must be adjacent (i.e. no intervening ($j1,$e1) with $j1 != $i1 or $e1 != $d1)
+##  + supports multiple lines for pairs ($i1,$d1,$i2) provided the above conditions hold
 ##  + supports loading of $cof->{N} from single-value lines
 ##  + %opts: clobber %$cof
 sub loadTextFh {
@@ -174,47 +207,59 @@ sub loadTextFh {
   $cof->logconfess("loadTextFh(): cannot load unopened database!") if (!$cof->opened);
 
   ##-- common variables
-  my $pack_f   = $cof->{pack_f};
-  my $pack_i   = $cof->{pack_i};
-  my $pack_r1  = "${pack_i}${pack_f}"; ##-- $r1 : [$end2,$f1] @ $i1
-  my $pack_r2  = "${pack_i}${pack_f}"; ##-- $r2 : [$i2,$f12]  @ end2($i1-1)..(end2($i1)-1)
-  my $len_r2   = packsize($pack_r2);
-  my $fmin     = $cof->{fmin} // 0;
-  my ($r1,$r2) = @$cof{qw(r1 r2)};
+  ##   $r1 : [$end2]            @ $i1
+  ##   $r2 : [$end3,$d1,$f1]*   @ end2($i1-1)..(end2($i1+1)-1)
+  ##   $r3 : [$i2,$f12]*        @ end3($d1-1)..(end3($d1+1)-1)
+  my $fmin  = $cof->{fmin} // 0;
+  my ($r1,$r2,$r3)                = @$cof{qw(r1 r2 r3)};
+  my ($pack_r1,$pack_r2,$pack_r3) = map {$_->{packas}} ($r1,$r2,$r3);
   $r1->truncate();
   $r2->truncate();
-  my ($fh1,$fh2) = ($r1->{fh},$r2->{fh});
+  $r3->truncate();
+  my ($fh1,$fh2,$fh3) = ($r1->{fh},$r2->{fh},$r3->{fh});
 
   ##-- iteration variables
-  my ($pos1,$pos2) = (0,0);
-  my ($i1_cur,$f1) = (-1,0);
-  my ($f12,$i1,$i2,$f);
+  my ($pos1,$pos2,$pos3) = (0,0,0);
+  my ($i1_cur,$d1_cur,$f1) = (-1,undef,0);
+  my ($f12,$i1,$d1,$i2,$d2,$f);
   my $N  = 0;	  ##-- total marginal frequency as extracted from %f12
   my $N1 = 0;     ##-- total N as extracted from single-element records
   my %f12 = qw(); ##-- ($i2=>$f12, ...) for $i1_cur
 
-  ##-- guts for inserting records from $i1_cur,%f12,$pos1,$pos2
+  ##-- guts for inserting records from $i1_cur,$d1_cur,%f12,$pos1,$pos2 : call on changed ($i1_cur,$d1_cur)
   my $insert = sub {
     if ($i1_cur >= 0) {
       if ($i1_cur != $pos1) {
 	##-- we've skipped one or more $i1 because it had no collocates (e.g. kern01 i1=287123="Untier/1906")
-	$fh1->print( pack($pack_r1,$pos2,0) x ($i1_cur-$pos1) );
+	$fh1->print( pack($pack_r1,$pos2) x ($i1_cur-$pos1) );
+	$pos1 = $i1_cur;
       }
-      ##-- dump r2-records for $i1_cur
+
+      ##-- dump r3-records for ($i1_cur,$d1_cur,*)
       $f1 = 0;
       foreach (sort {$a<=>$b} keys %f12) {
 	$f    = $f12{$_};
 	$f1  += $f;
-	next if ($f < $fmin); ##-- skip here so we can track "real" marginal frequencies
-	$fh2->print(pack($pack_r2, $_,$f));
+	next if ($f < $fmin || $_ < 0); ##-- skip here so we can track "real" marginal frequencies
+	$fh3->print(pack($pack_r3, $_,$f));
+	++$pos3;
+      }
+
+      ##-- dump r2-record for ($i1_cur,$d1_cur)
+      if (defined($d1_cur)) {
+	$fh2->print(pack($pack_r2, $pos3,$d1_cur,$f1));
 	++$pos2;
       }
-      ##-- dump r1-record for $i1_cur
-      $fh1->print(pack($pack_r1, $pos2,$f1));
-      $pos1  = $i1_cur+1;
-      $N    += $f1;
+
+      ##-- maybe dump r1-record for $i1_cur
+      if ($i1 != $i1_cur) {
+	$fh1->print(pack($pack_r1, $pos2));
+	$pos1  = $i1_cur+1;
+      }
+      $N += $f1;
     }
     $i1_cur = $i1;
+    $d1_cur = $d1;
     %f12    = qw();
   };
 
@@ -222,22 +267,30 @@ sub loadTextFh {
   binmode($infh,':raw');
   while (defined($_=<$infh>)) {
     chomp;
-    ($f12,$i1,$i2) = split(' ',$_,3);
+    ($f12,$i1,$d1,$i2,$d2) = split(' ',$_,5);
     if (!defined($i1)) {
       #$cof->debug("N1 += $f12");
       $N1 += $f12;		      ##-- load N values
       next;
     }
-    $insert->() if ($i1 != $i1_cur);  ##-- insert record(s) for $i1_cur
-    $f12{$i2} += $f12;                ##-- buffer co-frequencies for $i1_cur
+    elsif (!defined($d1)) {
+      $cof->logconfess("loadTextFh(): failed to parse input line ", $infh->input_line_number);
+    }
+    $insert->()			      ##-- insert record(s) for ($i1_cur,$d1_cur)
+      if ($i1 != $i1_cur || $d1 != $d1_cur);
+    $f12{$i2//-1} += $f12;            ##-- buffer co-frequencies for ($i1_cur,$d1_cur); track un-collocated frequencies as $i2=-1
   }
-  $insert->();                        ##-- write record(s) for final $i1_cur
+  $i1 = -1;
+  $insert->();                        ##-- write record(s) for final ($i1_cur,$d1_cur)
 
   ##-- adopt final $N and sizes
   #$cof->debug("FINAL: N1=$N1, N=$N");
   $cof->{N} = $N1>$N ? $N1 : $N;
-  $cof->{size1} = $r1->size;
-  $cof->{size2} = $r2->size;
+  foreach (qw(1 2 3)) {
+    my $r = $cof->{"r$_"};
+    $r->flush();
+    $cof->{"size$_"} = $r->size;
+  }
 
   return $cof;
 }
@@ -245,77 +298,10 @@ sub loadTextFh {
 ## $cof = $cof->loadTextFile_create($fh,%opts)
 ##  + old version of loadTextFile() which doesn't support N, semi-sorted input, or multiple ($i1,$i2) entries
 ##  + not useable by union() method
+##  + obsolete; now just an alias for loadTextFile()
 sub loadTextFile_create {
-  my ($cof,$infile,%opts) = @_;
-  my $infh = ref($infile) ? $infile : IO::File->new("<$infile");
-  if (!ref($cof)) {
-    $cof = $cof->new(%opts);
-  } else {
-    @$cof{keys %opts} = values %opts;
-  }
-  $cof->logconfess("loadTextFile_create(): cannot load unopened database!") if (!$cof->opened);
-
-  ##-- common variables
-  my $pack_f   = $cof->{pack_f};
-  my $pack_i   = $cof->{pack_i};
-  my $pack_r1  = "${pack_i}${pack_f}"; ##-- $r1 : [$end2,$f1] @ $i1
-  my $pack_r2  = "${pack_i}${pack_f}"; ##-- $r2 : [$i2,$f12]  @ end2($i1-1)..(end2($i1)-1)
-  my $len_r2   = packsize($pack_r2);
-  my $fmin     = $cof->{fmin} // 0;
-  my ($r1,$r2) = @$cof{qw(r1 r2)};
-  $r1->truncate();
-  $r2->truncate();
-  my ($fh1,$fh2) = ($r1->{fh},$r2->{fh});
-
-  ##-- iteration variables
-  my ($pos1,$pos2,$pos2_prev) = (0,0,0);
-  my ($i1_cur,$f1_cur) = (-1,0);
-  my ($f12,$i1,$i2);
-  my $N = 0;
-
-  ##-- guts for inserting records from $i1_cur,%f12,$pos1,$pos2
-  my $insert1 = sub {
-    if ($i1_cur >= 0) {
-      ##-- dump record for $i1_cur
-      if ($i1_cur != $pos1) {
-	##-- we've skipped one or more $i1 because it had no collocates (e.g. kern01 i1=287123="Untier/1906")
-	$fh1->print( pack($pack_r1,$pos2_prev,0) x ($i1_cur-$pos1) );
-      }
-      $fh1->print(pack($pack_r1, $pos2,$f1_cur));
-      $pos1      = $i1_cur+1;
-      $pos2_prev = $pos2;
-    }
-    $N     += $f1_cur;
-    $i1_cur = $i1;
-    $f1_cur = 0;
-  };
-
-  ##-- ye olde loope
-  binmode($infh,':raw');
-  while (defined($_=<$infh>)) {
-    ($f12,$i1,$i2) = split(' ',$_,3);
-    #next if ($f12 < $fmin);  		##-- don't skip here so that we can track "real" marginal frequencies
-    $insert1->() if ($i1 != $i1_cur);	##-- insert record for $i1_cur
-
-    ##-- track marginal f($i1)
-    $f1_cur += $f12;
-    next if ($f12 < $fmin		##-- minimum co-occurrence frequency filter
-	     #|| $i1==$i2  		##-- suppress identity collocations (... but we can't eliminate e.g. lemma-identity if using complex tuples!)
-	    );
-
-    ##-- dump record to $r2
-    $fh2->print(pack($pack_r2, $i2,$f12));
-    ++$pos2;
-  }
-  $insert1->(); 			##-- dump final record for $i1_cur
-
-  ##-- adopt final $N and sizes
-  $cof->{N} = $N;
-  $cof->{size1} = $r1->size;
-  $cof->{size2} = $r2->size;
-
-  $infh->close() if (!ref($infile));
-  return $cof;
+  my $cof = shift;
+  return $cof->loadTextFile(@_);
 }
 
 ## $bool = $obj->saveTextFile($filename_or_handle, %opts)
@@ -323,39 +309,55 @@ sub loadTextFile_create {
 ##  + INHERITED from DiaColloDB::Persistent
 
 ## $bool = $cof->saveTextFh($fh,%opts)
-##  + save from text file with lines of the form "N", "FREQ ID1 ID2"*
+##  + save from text file with lines of the form:
+##      N                 ##-- 1 field : N
+##      FREQ ID1 DATE     ##-- 3 fields: un-collocated portion of $f1
+##      FREQ ID1 DATE ID2 ##-- 4 fields: co-frequency pair (ID2 >= 0)
 ##  + %opts:
-##      i2s => \&CODE,   ##-- code-ref for formatting indices; called as $s=CODE($i)
+##      i2s => \&CODE,    ##-- code-ref for formatting indices; called as $s=CODE($i)
+##      i2s1 => \&CODE,   ##-- code-ref for formatting item1 indices (overrides 'i2s')
+##      i2s2 => \&CODE,   ##-- code-ref for formatting item2 indices (overrides 'i2s')
 sub saveTextFh {
   my ($cof,$outfh,%opts) = @_;
   $cof->logconfess("saveTextFile(): cannot save unopened DB") if (!$cof->opened);
 
   ##-- common variables
-  my ($r1,$r2)   = @$cof{qw(r1 r2)};
-  my $pack_r1    = $r1->{packas};
-  my $pack_r2    = $r2->{packas};
-  my $i2s        = $opts{i2s};
-  my ($fh1,$fh2) = ($r1->{fh},$r2->{fh});
+  ##   $r1 : [$end2]            @ $i1
+  ##   $r2 : [$end3,$d1,$f1]*   @ end2($i1-1)..(end2($i1+1)-1)
+  ##   $r3 : [$i2,$f12]*        @ end3($d1-1)..(end3($d1+1)-1)
+  my ($r1,$r2,$r3) = @$cof{qw(r1 r2 r3)};
+  my ($pack1,$pack2,$pack3) = map {$_->{packas}} ($r1,$r2,$r3);
+  my $i2s  = $opts{i2s};
+  my $i2s1 = exists($opts{i2s1}) ? $opts{i2s1} : $i2s;
+  my $i2s2 = exists($opts{i2s2}) ? $opts{i2s2} : $i2s;
 
   ##-- iteration variables
-  my ($buf1,$i1,$f1,$end2);
-  my ($buf2,$off2,$i2,$f12);
+  my ($buf1,$i1,$s1,$end2);
+  my ($buf2,$off2,$end3,$d1,$f1);
+  my ($buf3,$off3,$i2,$s2,$f12,$f12sum);
 
   ##-- ye olde loope
   binmode($outfh,':raw');
   $outfh->print($cof->{N}, "\n");
-  for ($r1->seek($i1=0), $r2->seek($off2=0); !$r1->eof(); ++$i1) {
+  for ($r1->seek($i1=0), $r2->seek($off2=0), $r3->seek($off3=0); !$r1->eof(); ++$i1) {
     $r1->read(\$buf1) or $cof->logconfess("saveTextFile(): failed to read record $i1 from $r1->{file}: $!");
-    ($end2,$f1) = unpack($pack_r1,$buf1);
+    $end2 = unpack($pack1,$buf1);
+    $s1 = $i2s1 ? $i2s1->($i1) : $i1;
 
     for ( ; $off2 < $end2 && !$r2->eof(); ++$off2) {
       $r2->read(\$buf2) or $cof->logconfess("saveTextFile(): failed to read record $off2 from $r2->{file}: $!");
-      ($i2,$f12) = unpack($pack_r2,$buf2);
-      $outfh->print($f12, "\t",
-		    ($i2s
-		     ? ($i2s->($i1), "\t", $i2s->($i2))
-		     : ($i1, "\t", $i2)),
-		    "\n");
+      ($end3,$d1,$f1) = unpack($pack2,$buf2);
+
+      for ($f12sum=0; $off3 < $end3 && !$r3->eof(); ++$off3) {
+	$r3->read(\$buf3) or $cof->logconfess("saveTextFile(): failed to read record $off3 from $r3->{file}: $!");
+	($i2,$f12) = unpack($pack3,$buf3);
+	$f12sum   += $f12;
+	$s2        = $i2s2 ? $i2s2->($i2) : $i2;
+	$outfh->print(join("\t", $f12, $s1, $d1, $s2), "\n");
+      }
+
+      ##-- track un-collocated portion of ($f1,$d1), if any
+      $outfh->print(join("\t", $f1-$f12sum, $s1, $d1), "\n") if ($f12sum != $f1);
     }
   }
 
@@ -367,11 +369,10 @@ sub saveTextFh {
 
 ## $rel = $CLASS_OR_OBJECT->create($coldb,$tokdat_file,%opts)
 ##  + populates current database from $tokdat_file,
-##    a tt-style text file containing 1 token-id perl line with optional blank lines
-##  + %opts: clobber %$ug, also:
-##    (
-##     size=>$size,  ##-- set initial size (number of types)
-##    )
+##    a tt-style text file with lines of the form:
+##      TID DATE	##-- single token
+##	"\n"		##-- blank line: EOS
+##  + %opts: clobber %$ug
 sub create {
   my ($cof,$coldb,$tokfile,%opts) = @_;
 
@@ -392,7 +393,7 @@ sub create {
   ##-- sort filter
   env_push(LC_ALL=>'C');
   my $tmpfile = "$cof->{base}.dat";
-  my $sortfh = opencmd("| sort -nk1 -nk2 | uniq -c - $tmpfile")
+  my $sortfh = opencmd("| sort -nk1 -nk2 -nk3 | uniq -c - $tmpfile")
     or $cof->logconfess("create(): open failed for pipe to sort|uniq: $!");
   binmode($sortfh,':raw');
 
@@ -425,7 +426,7 @@ sub create {
 
   ##-- stage2: load pair-frequencies
   $cof->vlog('trace', "create(): stage2: load pair frequencies (fmin=$cof->{fmin})");
-  $cof->loadTextFile_create($tmpfile)
+  $cof->loadTextFile($tmpfile)
     or $cof->logconfess("create(): failed to load pair frequencies from $tmpfile: $!");
 
   ##-- stage3: header
@@ -445,9 +446,9 @@ sub create {
 
 ## $cof = CLASS_OR_OBJECT->union($coldb, \@pairs, %opts)
 ##  + merge multiple unigram unigram indices from \@pairs into new object
-##  + @pairs : array of pairs ([$cof,\@xi2u],...)
-##    of unigram-objects $cof and tuple-id maps \@xi2u for $cof
-##    - \@xi2u may also be a mapping object supporting a toArray() method
+##  + @pairs : array of pairs ([$cof,\@ti2u],...)
+##    of unigram-objects $cof and tuple-id maps \@ti2u for $cof
+##    - \@ti2u may also be a mapping object supporting a toArray() method
 ##  + %opts: clobber %$cof
 ##  + implicitly flushes the new index
 sub union {
@@ -464,7 +465,7 @@ sub union {
   binmode($tmpfh,':raw');
 
   ##-- stage1: extract pairs and N
-  $cof->vlog('trace', "union(): stage1: extract pairs");
+  $cof->vlog('trace', "union(): stage1: collect pairs");
   my ($pair,$pcof,$pi2u);
   my $pairi=0;
   foreach $pair (@$pairs) {
@@ -477,14 +478,12 @@ sub union {
   $tmpfh->close()
     or $cof->logconfess("union(): failed to close tempfile $tmpfile: $!");
 
-  ##-- sort temp-file
+  ##-- stage2: sort & load tempfile
   env_push(LC_ALL=>'C');
-  my $sortfh = opencmd("sort -n -k2 -k3 $tmpfile |")
+  $cof->vlog('trace', "union(): stage2: load pair frequencies (fmin=$cof->{fmin})");
+  my $sortfh = opencmd("sort -n -k2 -k3 -k4 $tmpfile |")
     or $cof->logconfess("union(): open failed for pipe from sort: $!");
   binmode($sortfh,':raw');
-
-  ##-- stage2: load pair-frequencies
-  $cof->vlog('trace', "union(): stage2: load pair frequencies (fmin=$cof->{fmin})");
   $cof->loadTextFh($sortfh)
     or $cof->logconfess("union(): failed to load pair frequencies from $tmpfile: $!");
   $sortfh->close()
@@ -495,7 +494,7 @@ sub union {
   $cof->saveHeader()
     or $cof->logconfess("union(): failed to save header: $!");
 
-  ##-- unlink temp file
+  ##-- cleanup: unlink temp file
   CORE::unlink($tmpfile) if (!$cof->{keeptmp});
 
   return $cof;
@@ -509,164 +508,177 @@ sub union {
 sub dbinfo {
   my $cof = shift;
   my $info = $cof->SUPER::dbinfo();
-  @$info{qw(fmin dmax size1 size2 N)} = @$cof{qw(fmin dmax size1 size2 N)};
+  @$info{qw(fmin dmax size1 size2 size3 N)} = @$cof{qw(fmin dmax size1 size2 size3 N)};
   return $info;
 }
 
 
 ##==============================================================================
 ## Utilities: lookup
+##  + BROKEN in v0.10.000 (x(+date)->t(-date) db tuples)
 
-## $f = $cof->f1( @ids)
-## $f = $cof->f1(\@ids)
+## $f = $cof->f1( @xids)
+## $f = $cof->f1(\@xids)
 ##  + get total marginal unigram frequency (db must be opened)
+##  + no longer supported since v0.10.000
 sub f1 {
-  my $cof = shift;
-  my $ids = UNIVERSAL::isa($_[0],'ARRAY') ? @{$_[0]} : \@_;
-  my $r1  = $cof->{r1};
-  my $f   = 0;
-  foreach (@$ids) {
-    $f += $r1->fetch($_)->[1];
-  }
-  return $f;
+  $_[0]->logconfess("f1(): method no longer supported");
 }
 
-## $f12 = $cof->f12($id1,$id2)
-##  + return joint frequency for pair ($id1,$id2)
-##  + UNUSED
+## $f12 = $cof->f12($xid1,$xid2)
+##  + return joint frequency for pair ($xid1,$xid2)
+##  + no longer supported since v0.10.000
 sub f12 {
-  my ($cof,$i1,$i2) = @_;
-  my $beg2 = ($i1==0 ? 0 : $cof->{r1}->fetch($i1-1)->[0]);
-  my $end2 = $cof->{r1}->fetch($i1)->[0];
-  my $pos2 = $cof->{r2}->bsearch($i2, lo=>$beg2, hi=>$end2, packas=>$cof->{pack_i});
-  return defined($pos2) ? $cof->{r2}->fetch($pos2)->[1] : 0;
+  $_[0]->logconfess("f12(): method no longer supported");
 }
 
 ##==============================================================================
 ## Relation API: default: profiling
 
-## $prf = $cof->subprofile1(\@xids, %opts)
-##  + get joint co-frequency profile for @xids (db must be opened; f1 and f12 only)
-##  + %opts:
-##     groupby => \&gbsub,  ##-- key-extractor $key2_or_undef = $gbsub->($i2)
-##     coldb   => $coldb,   ##-- for debugging
-##     onepass => $bool,    ##-- use fast but incorrect 1-pass method?
+## \%slice2prf = $rel->subprofile1(\@tids,\%opts)
+##  + get slice-wise joint co-frequency profile(s) for @tids (db must be opened; f1 and f12 only)
+##  + %opts: as for profile(), also:
+##     coldb => $coldb,   ##-- parent DiaColloDB object (for shared data, debugging)
+##     dreq  => \%dreq,   ##-- parsed date request
 sub subprofile1 {
-  my ($cof,$ids,%opts) = @_;
-  $ids   = [$ids] if (!UNIVERSAL::isa($ids,'ARRAY'));
-  my $r1 = $cof->{r1};
-  my $r2 = $cof->{r2};
-  my $pack1 = $r1->{packas};
-  my $pack2 = $r2->{packas};
-  my $pack1i = $cof->{pack_i};
-  my $pack1f = "@".packsize($cof->{pack_i}).$cof->{pack_f};
+  my ($cof,$tids,$opts) = @_;
+
+  ##-- common variables
+  $tids = [$tids] if (!UNIVERSAL::isa($tids,'ARRAY'));
+  my $coldb = $opts->{coldb};
+  my $slice = $opts->{slice};
+  my $dreq  = $opts->{dreq};
+  my $dfilter = $dreq->{dfilter};
+  my $groupby = $opts->{groupby}{ti2g};
+  my $onepass = $opts->{onepass};
+  my $pack_id = $coldb->{pack_id};
+
+  ##-- vars: relation-wise
+  ##   $r1 : [$end2]            @ $i1
+  ##   $r2 : [$end3,$d1,$f1]*   @ end2($i1-1)..(end2($i1+1)-1)
+  ##   $r3 : [$i2,$f12]*        @ end3($d1-1)..(end3($d1+1)-1)
+  my ($r1,$r2,$r3)          = @$cof{qw(r1 r2 r3)};
+  my ($pack1,$pack2,$pack3) = map {$_->{packas}} ($r1,$r2,$r3);
+  my $pack2e = $cof->{pack_i};
+  my $pack2d = '@'.packsize("$cof->{pack_i}").$cof->{pack_d};
+  my $pack2f = '@'.packsize("$cof->{pack_i}$cof->{pack_d}").$cof->{pack_f};
   my $size1  = $cof->{size1} // ($cof->{size1}=$r1->size);
   my $size2  = $cof->{size2} // ($cof->{size2}=$r2->size);
-  my $groupby = $opts{groupby};
-  my $pack_id = $opts{coldb}{pack_id};
-  my $onepass = $opts{onepass};
-  my $pf1 = 0;
-  my $pf12 = {};
-  my $pf2  = {};
-  my ($i1,$i2,$key2, $beg2,$end2,$pos2, $f1,$f12, $buf, %i2);
+  my $size3  = $cof->{size3} // ($cof->{size3}=$r3->size);
 
-  foreach $i1 (@$ids) {
+  ##-- setup %slice2prf
+  my %slice2prf = map {
+    ($_ => DiaColloDB::Profile->new(f1=>0, N=>$cof->{N}))
+  } ($slice ? (map {$_*$slice} (($dreq->{slo}/$slice)..($dreq->{shi}/$slice))) : 0);
+
+  ##-- ye olde loope
+  my ($i1,$beg2,$end2, $pos2,$beg3,$end3,$d1,$ds,$dprf,$f1, $pos3,$i2,$f12,$key2, $buf,%id2);
+  my ($blo,$bhi,$bi); ##-- one-pass guts
+  foreach $i1 (@$tids) {
     next if ($i1 >= $size1);
-    $beg2       = ($i1==0 ? 0 : unpack($pack1i,$r1->fetchraw($i1-1,\$buf)));
-    ($end2,$f1) = unpack($pack1, $r1->fetchraw($i1,\$buf));
+    $beg2 = ($i1==0 ? 0 : unpack($pack1,$r1->fetchraw($i1-1,\$buf)));
+    $end2 = unpack($pack1, $r1->fetchraw($i1,\$buf));
 
-    $pf1       += $f1;
     next if ($beg2 >= $size2);
-    for ($r2->seek($beg2), $pos2=$beg2; $pos2 < $end2; ++$pos2) {
-      $r2->getraw(\$buf) or last;
-      ($i2,$f12)    = unpack($pack2, $buf);
-      $key2         = $groupby ? $groupby->($i2) : pack($pack_id,$i2);
-      next if (!defined($key2)); ##-- item2 selection via groupby CODE-ref
-      $pf12->{$key2} += $f12;
-      if ($onepass && !exists($i2{$i2})) {
-	$pf2->{$key2} += unpack($pack1f, $r1->fetchraw($i2,\$buf)); ##-- avoid double-counting f2 for shared collocates
-	$i2{$i2}       = undef;
+    for ($pos2=$beg2; $pos2 < $end2; ++$pos2) {
+      $beg3           = ($pos2==0 ? 0 : unpack($pack2e, $r2->fetchraw($pos2-1,\$buf)));
+      ($end3,$d1,$f1) = unpack($pack2, $r2->fetchraw($pos2,\$buf));
+
+      ##-- check date-filter & get slice-local profile $dprf
+      next if ($dfilter && !$dfilter->($d1));
+      $ds   = $slice ? int($d1/$slice)*$slice : 0;
+      $dprf = $slice2prf{$ds};
+      $dprf->{f1} += $f1;
+
+      next if ($beg3 >= $size3);
+      for ($pos3=$beg3; $pos3 < $end3; ++$pos3) {
+	($i2,$f12) = unpack($pack3, $r3->fetchraw($pos3,\$buf));
+	$key2      = $groupby ? $groupby->($i2) : pack($pack_id,$i2);
+	next if (!defined($key2)); ##-- item2 selection via groupby CODE-ref
+	$dprf->{f12}{$key2} += $f12;
+
+	if ($onepass && !exists($id2{"$i2 $d1"})) {
+	  ##-- search for ($i2,$date) offset in r2
+	  $id2{"$i2 $d1"} = undef;
+	  $blo = ($i2==0 ? 0 : unpack($pack1,$r1->fetchraw($i2-1,\$buf)));
+	  $bhi = unpack($pack1, $r1->fetchraw($i2,\$buf));
+	  $bi  = $r2->bsearch($d1,lo=>$blo,hi=>$bhi,packas=>$pack2d);
+	  $dprf->{f2}{$key2} += unpack($pack2f, $r2->fetchraw($bi,\$buf));
+	}
       }
     }
   }
-  return DiaColloDB::Profile->new(
-				  N=>$cof->{N},
-				  f1=>$pf1,
-				  f2=>$pf2,
-				  f12=>$pf12,
-				 );
+
+  return \%slice2prf;
 }
 
-##  \%slice2prf = $rel->subprofile2(\%slice2prf, %opts)
+##  \%slice2prf = $rel->subprofile2(\%slice2prf, \%opts)
 ##  + populate f2 frequencies for profiles in \%slice2prf
-##  + %opts:
-##     groupby => \%gbreq,  ##-- parsed groupby object
-##     a2data  => \%a2data, ##-- maps indexed attributes to associated datastructures
-##     coldb   => $coldb,   ##-- parent DiaColloDB object (for shared data, debugging)
-##     #slices  => \@slices, ##-- slices (optional)
-##     ...                  ##-- other options as for profile(), esp. qw(slice)
-##  + default implementation does nothing
+##  + %opts: as for subprofile1()
 sub subprofile2 {
-  my ($cof,$slice2prf,%opts) = @_;
+  my ($cof,$slice2prf,$opts) = @_;
 
   ##-- vars: common
-  my $coldb   = $opts{coldb};
-  my $groupby = $opts{groupby};
-  my $a2data  = $opts{a2data};
-  my $slice   = $opts{slice};
-  #my $slices  = $opts{slices} || [sort {$a<=>$b} keys %$slice2prf];
-  my ($dfilter,$slo,$shi,$dlo,$dhi) = $coldb->parseDateRequest(@opts{qw(date slice fill)});
-  my $filter_by_date = $slice || defined($dlo) || defined($dhi);
+  my $coldb   = $opts->{coldb};
+  my $groupby = $opts->{groupby};
+  my $a2data  = $opts->{a2data};
+  my $slice   = $opts->{slice};
+  my $dfilter = $opts->{dreq}{dfilter};
 
   ##-- vars: relation-wise
-  my $r1       = $cof->{r1};
-  my $pack_r1f = '@'.packsize($cof->{pack_i}).$cof->{pack_f};
+  ##   $r1 : [$end2]            @ $i1
+  ##   $r2 : [$end3,$d1,$f1]*   @ end2($i1-1)..(end2($i1+1)-1)
+  ##   #$r3 : [$i2,$f12]*        @ end3($d1-1)..(end3($d1+1)-1)
+  my ($r1,$r2)  = @$cof{qw(r1 r2)};
+  my $pack1   = $r1->{packas};
+  my $pack2df = '@'.packsize("$cof->{pack_i}",0)."$cof->{pack_d}$cof->{pack_i}";
+
+  ##-- optimize tightest loop for direct mmap buffer access if available
+  my $bufr2 = UNIVERSAL::isa($r2,'DiaColloDB::PackedFile::MMap') ? $r2->{bufr} : undef;
+  my $len2  = $r2->{reclen};
 
   ##-- get "most specific projected attribute" ("MSPA"): that projected attribute with largest enum
   #my $gb1      = scalar(@{$groupby->{attrs}})==1; ##-- are we grouping by a single attribute? -->optimize!
   my $mspai    = (sort {$b->[1]<=>$a->[1]} map {[$_,$a2data->{$groupby->{attrs}[$_]}{enum}->size]} (0..$#{$groupby->{attrs}}))[0][0];
   my $mspa     = $groupby->{attrs}[$mspai];
   my $mspgpack = $groupby->{gpack}[$mspai];
-  my $mspxpack = $groupby->{xpack}[$mspai];
-  my $msp2x = $a2data->{$mspa}{a2x};
-  my %mspv  = qw(); ##-- checked MSPA-values ($mspvi)
-  my $xenum = $coldb->{xenum};
-  my $pack_xd = "@".(packsize($coldb->{pack_id}) * scalar(@{$coldb->{attrs}})).$coldb->{pack_date};
-  my $xs2g    = $groupby->{xs2g};
-  my $debug_xp2g = join('',@{$groupby->{xpack}});
-  my $debug_gpack= "($coldb->{pack_id})*";
-  my ($prf,$pf12, $mspvi,$i2,$x2,$d2,$ds2,$prf2,$key2, $buf,$f2);
-  $prf2     = (values %$slice2prf)[0] if (!$filter_by_date);
-  foreach $prf (values %$slice2prf) {
-    $pf12 = $prf->{f12};
-    foreach (keys %$pf12) {
+  my $msptpack = $groupby->{tpack}[$mspai];
+  my $msp2t    = $a2data->{$mspa}{a2t};
+  my %mspv     = qw(); ##-- checked MSPA-values ($mspvi)
+  my $tenum    = $coldb->{tenum};
+  my $ts2g     = $groupby->{ts2g};
+
+  my ($prf1, $mspvi,$i2,$t2,$key2, $beg2,$end2,$pos2, $d2,$f2,$ds2,$prf2, $buf);
+  foreach $prf1 (values %$slice2prf) {
+    foreach (keys %{$prf1->{f12}}) {
       $mspvi = unpack($mspgpack,$_);
       next if (exists $mspv{$mspvi});
       $mspv{$mspvi} = undef;
-      foreach $i2 (@{$msp2x->fetch($mspvi)}) {
-	##-- get item2 x-tuple
-	$x2  = $xenum->i2s($i2);
 
-	if ($filter_by_date) {
-	  ##-- extract item2 date slice
-	  $d2  = unpack($pack_xd, $x2);
+      foreach $i2 (@{$msp2t->fetch($mspvi)}) {
+	##-- get item2 t-tuple
+	$t2 = $tenum->i2s($i2);
+
+	##-- get groupby-key from tuple-string
+	next if (!defined($key2 = $ts2g ? $ts2g->($t2) : pack($mspgpack, $i2))); ##-- having() failure
+
+	##-- scan all dates for $i2
+	$beg2 = ($i2==0 ? 0 : unpack($pack1,$r1->fetchraw($i2-1,\$buf)));
+	$end2 = unpack($pack1, $r1->fetchraw($i2,\$buf));
+	for ($pos2=$beg2; $pos2 < $end2; ++$pos2) {
+	  ($d2,$f2) = unpack($pack2df, $bufr2 ? substr($$bufr2, $pos2*$len2, $len2) : $r2->fetchraw($pos2,\$buf));
+
+	  ##-- check date-filter & get slice
+	  next if ($dfilter && !$dfilter->($d2));
 	  $ds2 = $slice ? int($d2/$slice)*$slice : 0;
 
-	  ##-- ignore if item2 slice isn't in our target range
-	  next if (!defined($prf2=$slice2prf->{$ds2})
-		   || (defined($dlo) && $d2 < $dlo)
-		   || (defined($dhi) && $d2 > $dhi));
+	  ##-- ignore if item2 isn't in target slice
+	  $prf2 = $slice2prf->{$ds2};
+	  next if (!exists($prf2->{f12}{$key2}));
+
+	  ##-- add independent f2
+	  $prf2->{f2}{$key2} += $f2;
 	}
-
-	##-- get groupby-key from x-tuple string & check for item2 membership in the appropriate slice-profile
-	$key2 = $xs2g ? $xs2g->($x2) : pack($mspgpack, $i2);
-	#$key2 = pack($debug_gpack, unpack($debug_xp2g, $x2)); ##-- ca. 6% faster than $xs2g, no having-checks
-	#$key2 = pack($debug_gpack, $mspvi); ##-- ca. 12% faster than $xs2g, no having-checks, only valid for groupby single-attribute
-	next if (!defined($key2) || !exists($prf2->{f12}{$key2})); ##-- having()-failure or no item2 in target slice
-
-	##-- add item2 frequency
-	$f2 = unpack($pack_r1f, $r1->fetchraw($i2,\$buf));
-	$prf2->{f2}{$key2} += $f2;
       }
     }
   }
